import json
import select
import sys

from typing import Any, Optional
from django.core.management.base import BaseCommand

from argparse import ArgumentParser
from domain_event_broker import replay


def input_timeout(timeout: int) -> Optional[str]:
    inp, _, _ = select.select([sys.stdin], [], [], timeout)
    return sys.stdin.readline().strip() if inp else None


class Command(BaseCommand):

    help = "Move dead-lettered event back into given subscriber queue"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument('queue', nargs='+', type=str)
        parser.add_argument(
            '--all',
            action='store_true',
            dest='replay_all',
            default=False,
            help='Replay all events in dead letter queue.',
        )
        parser.add_argument(
            '--interactive',
            action='store_true',
            dest='interactive',
            default=False,
            help='Ask for desired action for each event.',
        )

    def interactive_filter(self, body: bytes, **kwargs: Any) -> str:
        payload = json.loads(body)
        self.stdout.write("Please specify action for:")
        self.stdout.write(json.dumps(payload, indent=4, sort_keys=True))
        self.stdout.write("(R)eplay, (D)iscard or (L)eave? ", ending="")
        # Wait for user input but time out before the RabbitMQ connection is
        # lost.
        action = input_timeout(30)
        if action is None:
            self.stdout.write("Timed out. Message left in queue.")
            return replay.LEAVE
        elif action == '':
            self.stdout.write("Message left in queue.")
            return replay.LEAVE
        action = action[0].upper()
        if action == 'R':
            self.stdout.write("Message is returned to worker queue.")
            return replay.RETRY
        elif action == 'D':
            self.stdout.write("Message has been discarded.")
            return replay.DISCARD
        else:
            self.stdout.write("Message left in queue.")
            return replay.LEAVE

    def handle(self, *args: Any, **options: Any) -> None:
        callback = replay.retry_event
        if options['interactive']:
            callback = self.interactive_filter
        for queue in options['queue']:
            if options['replay_all']:
                remaining = replay.replay_all(queue, callback)
            else:
                remaining = replay.replay_event(queue, callback)
            self.stdout.write("{} dead-lettered events remaining for {}".format(remaining, queue))



import pytest
from unittest.mock import patch
from django.db import transaction

from domain_event_broker.django import publish_on_commit


@pytest.mark.django_db(transaction=True)
@patch('domain_event_broker.django.transaction.publish_domain_event')
def test_publish_no_transaction(publish_mock):
    publish_on_commit('test.no_transaction', {})
    publish_mock.assert_called_with('test.no_transaction', {})


@pytest.mark.django_db(transaction=True)
@patch('domain_event_broker.django.transaction.publish_domain_event')
def test_publish_on_commit(publish_mock):
    with transaction.atomic():
        publish_on_commit('test.commit', {})
        assert not publish_mock.called
    publish_mock.assert_called_with('test.commit', {})


class RollbackError(Exception):
    pass


@pytest.mark.django_db(transaction=True)
@patch('domain_event_broker.django.transaction.publish_domain_event')
def test_discard_on_rollback(publish_mock):
    try:
        with transaction.atomic():
            publish_on_commit('test.rollback', {})
            raise RollbackError('Rollback transaction')
    except RollbackError:
        pass
    assert not publish_mock.called
    # Also don't emit event in subsequent transaction
    with transaction.atomic():
        pass
    assert not publish_mock.called



from domain_event_broker import publish_domain_event, Subscriber
from .helpers import check_queue_exists, get_queue_size
import uuid


def test_publish():
    def handle_event(event):
        handle_event.message = event.data['message']
    handle_event.message = None
    name = 'test-publish'
    subscriber = Subscriber()
    subscriber.register(handle_event, name, ['test.publish'])
    data = dict(message=str(uuid.uuid4())[:4])
    publish_domain_event('test.publish', data)
    subscriber.start_consuming(timeout=1.0)
    assert handle_event.message == data['message']
    assert get_queue_size(name) == 0


def test_dummy_mode():
    def handle_event(event):
        handle_event.message = event.data['message']
    handle_event.message = None
    name = 'test-dummy'
    subscriber = Subscriber(connection_settings=None)
    subscriber.register(handle_event, name, ['test.publish-dummy'])
    publish_domain_event('test.publish-dummy', {}, connection_settings=None)
    subscriber.start_consuming(timeout=1.0)
    assert not check_queue_exists(name)



from ._version import get_versions  # type: ignore
__version__ = get_versions()['version']
del get_versions

from .transport import (
    publish_domain_event,
    Subscriber,
    Retry,
    Publisher,
)

from .replay import (
    replay_event,
    replay_all,
    RETRY,
    DISCARD,
    LEAVE,
)

from .events import (
    DomainEvent,
)



from io import StringIO
from time import sleep
from django.core.management import call_command
from ..helpers import get_message_from_queue, get_queue_size


def test_replay(dead_letter_message):
    call_command('replay_domain_event', 'test-replay')
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 0
    header, event = get_message_from_queue('test-replay')
    assert event.data == dead_letter_message


def test_interactive_replay(dead_letter_message, monkeypatch):
    monkeypatch.setattr('sys.stdin', StringIO('Replay'))
    monkeypatch.setattr('select.select', lambda x, y, z, *args: (x, y, z))
    output = StringIO()
    call_command('replay_domain_event', 'test-replay', '--interactive',
                 stdout=output)
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 0
    header, event = get_message_from_queue('test-replay')
    assert event.data == dead_letter_message
    # Check JSON body is pretty-printed and sorted
    output_lines = output.getvalue().split('\n')
    assert output_lines[:3] == [
        'Please specify action for:',
        '{',
        '    "data": {'
    ]


def test_interactive_replay_leave(dead_letter_message, monkeypatch):
    monkeypatch.setattr('sys.stdin', StringIO('Leave'))
    monkeypatch.setattr('select.select', lambda x, y, z, *args: (x, y, z))
    call_command('replay_domain_event', 'test-replay', '--interactive')
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 1
    assert get_queue_size('test-replay') == 0


def test_interactive_replay_discard(dead_letter_message, monkeypatch):
    monkeypatch.setattr('sys.stdin', StringIO('Discard'))
    monkeypatch.setattr('select.select', lambda x, y, z, *args: (x, y, z))
    call_command('replay_domain_event', 'test-replay', '--interactive')
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 0
    assert get_queue_size('test-replay') == 0


def test_interactive_replay_invalid(dead_letter_message, monkeypatch):
    monkeypatch.setattr('sys.stdin', StringIO('Obliterate'))
    monkeypatch.setattr('select.select', lambda x, y, z, *args: (x, y, z))
    call_command('replay_domain_event', 'test-replay', '--interactive')
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 1
    assert get_queue_size('test-replay') == 0


def test_interactive_replay_timeout(dead_letter_message, monkeypatch):
    monkeypatch.setattr('sys.stdin', StringIO(''))
    monkeypatch.setattr('select.select', lambda x, y, z, *args: ([], [], []))
    call_command('replay_domain_event', 'test-replay', '--interactive')
    sleep(.6)
    assert get_queue_size('test-replay-dl') == 1
    assert get_queue_size('test-replay') == 0



from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import logging
import json
from pika import (
    BasicProperties,
    BlockingConnection,
    URLParameters,
    )
from pika import channel, frame, spec
from .events import DomainEvent
from . import settings

log = logging.getLogger(__name__)


def publish_domain_event(routing_key: str,
                         data: Dict[str, Any],
                         domain_object_id: str = None,
                         uuid_string: Optional[str] = None,
                         timestamp: Optional[float] = None,
                         connection_settings: Optional[str] = '',
                         ) -> DomainEvent:
    """
    Send a domain event to the message broker. The broker will take care of
    dispatching the event to registered subscribers.

    :param str routing_key: The routing key is of the form
        ``<DOMAIN>.<EVENT_TYPE>``.  The routing key should be a descriptive
        name of the domain event such as ``user.registered``.
    :param dict data: The actual event data. *Must* be json serializable.
    :param str domain_object_id: Domain identifier of the event. This field
        is optional. If used, it might make search in an event store easier.
    :param str uuid_string: This UUID identifier of the event. If left
        ``None``, a new one will be created.
    :param float timestamp: Unix timestamp. If timestamp is None, a new
        (UTC) timestamp will be created.
    :param str connection_settings: Specify the broker with an AMQP URL. If not
        given, the default broker will be used. If set to ``None``, the domain
        event is not published to a broker.
    :return: The domain event that was published.
    :rtype: :py:class:`domain_event_broker.DomainEvent`
    """
    event = DomainEvent(
        routing_key=routing_key,
        data=data,
        domain_object_id=domain_object_id,
        uuid_string=uuid_string,
        timestamp=timestamp)
    json_data = json.dumps(event.event_data)
    if connection_settings == '':
        connection_settings = settings.BROKER
    publisher = Publisher(connection_settings)
    publisher.publish(json_data, event.routing_key)
    publisher.disconnect()
    return event


class Retry(Exception):
    """
    Raise this exception in an event handler to schedule a delayed retry. The
    delay is specified in seconds.

    .. note::

        Internally a delay exchange with a per-message TTL is used and all
        delayed events for a handler that share the same delay are placed in
        one queue. The RabbitMQ TTL has a Millisecond resolution.
    """
    def __init__(self, delay: float = 10.0):
        super(Retry, self).__init__()
        self.delay = delay


def _retry_message(name: str,
                   retry_exchange: str,
                   channel: channel.Channel,
                   method: frame.Method,
                   properties: spec.BasicProperties,
                   body: str,
                   delay: float,
                   ) -> None:
    delay = int(delay * 1000)
    # Create queue that should be automatically deleted shortly after
    # the last message expires. The queue is re-declared for each retry
    # which resets the queue expiry.
    delay_name = '{}-delay-{}'.format(name, delay)
    result = channel.queue_declare(
        queue=delay_name,
        durable=True,
        arguments={
            'x-dead-letter-exchange': retry_exchange,
            'x-message-ttl': delay,
            'x-expires': delay + 10000,
            },
        )
    queue_name = result.method.queue
    # Bind the wait queue to the delay exchange before publishing
    channel.exchange_declare(
        exchange=delay_name,
        durable=True,
        auto_delete=True,  # Delete exchange when queue is deleted
        exchange_type='topic')
    channel.queue_bind(
        exchange=delay_name,
        routing_key='#',
        queue=queue_name)
    channel.basic_publish(
        exchange=delay_name,
        routing_key=method.routing_key,
        body=body,
        properties=properties)


def _call_event_handler(handler: Callable,
                        event: DomainEvent,
                        connection: BlockingConnection,
                        acknowledge: Callable,
                        retry: Callable,
                        reject: Callable,
                        max_retries: int,
                        ) -> None:
    # The handler is executed in a separate worker thread. Handle any errors
    # and trigger retries, dead-lettering or acknowledgement via threadsafe
    # callback on the connection.
    try:
        handler(event)
    except Retry as error:
        if event.retries < max_retries:
            # Publish manually to the delay exchange with a per-message TTL
            msg = "Retry ({retries}) consuming event {event} in {delay:.1f}s"
            log.info(msg.format(
                event=event,
                retries=event.retries,
                delay=error.delay))
            delayed_retry = partial(retry, delay=error.delay)
            connection.add_callback_threadsafe(acknowledge)
            connection.add_callback_threadsafe(delayed_retry)
        else:
            # Reject puts the message into the dead-letter queue if there is
            # one, otherwise the message is discarded.
            msg = "Exceeded max retries ({}) for {} event".format(
                max_retries, event.routing_key)
            log.error(msg, exc_info=True, extra=event.event_data)
            connection.add_callback_threadsafe(reject)
    except:  # noqa: E722
        # Note: If we want immediate requeueing, add a `RequeueError`
        # that a consumer can raise to trigger requeuing. Dead-letter
        # queues are a better choice in most cases.
        connection.add_callback_threadsafe(reject)
        log.exception("Event has been dead-lettered or discarded")
    else:
        connection.add_callback_threadsafe(acknowledge)


def receive_callback(transport: 'Subscriber',
                     handler: Callable,
                     name: str,
                     retry_exchange: str,
                     max_retries: int,
                     channel: channel.Channel,
                     method: frame.Method,
                     properties: spec.BasicProperties,
                     body: str,
                     ) -> None:
    try:
        event = DomainEvent.from_json(body)
    except Exception:
        # We cannot parse the message; requeuing would not help.
        channel.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
        log.exception("Failed to load message: %s", body)
    else:
        if properties.headers and 'x-death' in properties.headers:
            # Older RabbitMQ versions (< 3.5) keep adding x-death entries, new
            # versions only keep the most recent entry and increment 'count',
            # see: https://github.com/rabbitmq/rabbitmq-server/issues/78
            expiry_info = properties.headers['x-death'][0]
            if 'count' in expiry_info:
                event.retries = expiry_info['count']
            else:
                event.retries = len(properties.headers['x-death'])
        log.debug("Received {}:{}".format(method.routing_key, event))

        # The channel and connection objects are not threadsafe. Only call any
        # of those function from the main thread via a threadsafe callback.
        acknowledge = partial(
            channel.basic_ack,
            delivery_tag=method.delivery_tag)
        reject = partial(
            channel.basic_reject,
            delivery_tag=method.delivery_tag,
            requeue=False)
        retry = partial(
            _retry_message,
            name=name,
            retry_exchange=retry_exchange,
            channel=channel,
            method=method,
            properties=properties,
            body=body)
        event_handler = partial(
            _call_event_handler,
            handler=handler,
            event=event,
            connection=channel.connection,
            acknowledge=acknowledge,
            retry=retry,
            reject=reject,
            max_retries=max_retries)
        transport.workers.submit(event_handler)


def requires_broker(method: Callable) -> Callable:
    """
    If connection_settings are set to ``None`` on the transport object, don't
    perform the action and log the call instead. This is used for environments
    where no broker is available, e.g. development and testing.
    """
    def wrapper(transport: 'Transport', *args: Any, **kwargs: Any) -> Any:
        if transport.connection_settings is None:
            log.debug("No broker configured: {}.{}() is deactivated.".format(
                transport.__class__.__name__,
                method.__name__))
        else:
            return method(transport, *args, **kwargs)
    return wrapper


class Transport(object):

    def __init__(self,
                 connection_settings: Optional[str] = '',
                 exchange: str = "domain-events",
                 exchange_type: str = "topic",
                 ):
        if connection_settings == '':
            connection_settings = settings.BROKER
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.context_depth = 0
        self.connection = None
        self.connection_settings = connection_settings
        self.channel = None
        self.connect()

    @requires_broker
    def connect(self) -> None:
        """
        For now we use a synchronous connection - caller is blocked until a
        message is added to the queue. We might switch to asynch connections
        should this incur noticable latencies.
        """
        params = URLParameters(self.connection_settings)
        connection = BlockingConnection(params)
        channel = connection.channel()

        # set up the Exchange (if it does not exist)
        channel.exchange_declare(
            exchange=self.exchange,
            exchange_type=self.exchange_type,
            durable=True,
            auto_delete=False)

        self.connection = connection
        self.channel = channel

    @requires_broker
    def disconnect(self) -> None:
        """
        Disconnect from queue. The API is a little weird. First we close the
        connection, then disconnect from the socket. The last step will remove
        the connection from the RabbitMQ connection pool.

        Make sure you don't close the channel, otherwise the connection cannot
        be closed anymore.
        """
        if self.connection is not None:
            self.connection.close()
        self.channel = None
        self.connection = None


class Publisher(Transport):

    @requires_broker
    def publish(self, message: bytes, routing_key: Optional[str] = None) -> None:
        """
        Send as persistent message.
        """
        if self.channel is None:
            raise Exception('Not connected to broker.')

        self.channel.basic_publish(
            exchange=self.exchange,
            routing_key=routing_key,
            body=message,
            properties=BasicProperties(delivery_mode=2),
            )


class Subscriber(Transport):
    """
    A subscriber manages the registration of one or more event handlers. Once
    instantiated, call ``register`` to add subscribers and ``start_consuming``
    to wait for incoming events.

    .. note::

        The subscriber only uses one thread for processing events. Even if
        multiple handlers are registered, only one event is processed at a
        time.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.workers = ThreadPoolExecutor(max_workers=1)

    @requires_broker
    def bind_routing_keys(self,
                          exchange: str,
                          queue_name: str,
                          binding_keys: Union[List[str], Tuple[str]],
                          ) -> None:
        if self.channel is None:
            raise Exception('Not connected to broker.')

        for binding_key in binding_keys:
            self.channel.queue_bind(
                exchange=exchange,
                routing_key=binding_key,
                queue=queue_name)

    @requires_broker
    def register(self,
                 handler: Callable,
                 name: str,
                 binding_keys: Union[List[str], Tuple[str]],
                 dead_letter: bool = False,
                 durable: bool = True,
                 exclusive: bool = False,
                 auto_delete: bool = False,
                 max_retries: int = 0,
                 ) -> None:
        """
        Register a handler for one or more types of domain events.

        :param function handler: This function will be called when an event
            happens. It receives a ``DomainEvent`` as the only parameter.
        :param str name: Name of the handler. Used as the queue name.
        :param tuple|list binding_keys: One or more routing keys, e.g.
            ``["user.registered", "user.imported"]``. The binding keys may
            include wildcards. Use ``user.*`` to subscribe to all events in the
            user domain. Use ``#`` to call ``handler`` for all domain events
            across all domains.
        :param bool dead_letter: Whether to store events in a dead-letter queue
            if the handler raises an exception while processing the event.
        :param int max_retries: The handler may raise ``domain_event_broker.Retry``
            to indicate the event processing should be retried later. This
            parameter controls how often an event is rescheduled before it is
            dead-lettered or discarded.
        """
        if self.channel is None:
            raise Exception('Not connected to broker.')

        retry_exchange = name + '-retry'
        dead_letter_exchange = name + '-dlx'

        arguments = {}
        if dead_letter:
            self.channel.exchange_declare(
                exchange=dead_letter_exchange,
                exchange_type=self.exchange_type)
            result = self.channel.queue_declare(
                queue=name + '-dl',
                durable=True)
            queue_name = result.method.queue
            self.bind_routing_keys(
                dead_letter_exchange,
                queue_name,
                binding_keys)
            arguments["x-dead-letter-exchange"] = dead_letter_exchange

        # Create subscriber queue and bind to the default exchange
        self.channel.queue_declare(
            queue=name,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            arguments=arguments)
        self.bind_routing_keys(self.exchange, name, binding_keys)

        # Re-route failed messages to a retry dead letter queue.
        # This is only used if max_retries > 0 but we set up the exchanges
        # anyway so that messages can be replayed manually in the consumer
        # context.
        # Declare the exchange where messages expired in the wait queue are
        # routed.
        self.channel.exchange_declare(
            exchange=retry_exchange,
            exchange_type=self.exchange_type)
        # Bind the consumer queue to the retry exchange
        self.bind_routing_keys(retry_exchange, name, binding_keys)

        callback = partial(
            receive_callback,
            self,
            handler,
            name,
            retry_exchange,
            max_retries)
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(queue=name, on_message_callback=callback)

    @requires_broker
    def stop_consuming(self) -> None:
        if self.channel is not None:
            self.channel.stop_consuming()
        self.disconnect()

    @requires_broker
    def start_consuming(self, timeout: Optional[float] = None) -> None:
        """
        Enter IO consumer loop after calling `register`. If timeout is given,
        the consumer will be stopped after the specified number of seconds.
        """
        if self.connection is None or self.channel is None:
            raise Exception('Not connected to broker.')
        if timeout:
            self.connection.call_later(timeout, self.stop_consuming)
        try:
            self.channel.start_consuming()
        except KeyboardInterrupt:
            self.stop_consuming()
        except:  # noqa: E722
            self.stop_consuming()
            raise



import versioneer

from setuptools import setup, find_packages

# Add README as description
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='domain-event-broker',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Send and receive domain events via RabbitMQ',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ableton AG',
    author_email='webteam@ableton.com',
    url='https://github.com/AbletonAG/domain-event-broker',
    license='MIT',
    packages=find_packages(),
    package_data={'domain_event_broker': ['py.typed']},
    install_requires=["pika >= 1.2.0"],
    tests_require=["pytest >= 3.6.0"],
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from domain_event_broker import replay_event, Subscriber, DISCARD, LEAVE
from .helpers import get_message_from_queue, get_queue_size


def test_replay(dead_letter_message):
    assert replay_event('test-replay') == 0
    assert get_queue_size('test-replay-dl') == 0
    header, event = get_message_from_queue('test-replay')
    assert event.data == dead_letter_message


def discard(**kwargs):
    return DISCARD


def test_discard(dead_letter_message):
    assert replay_event('test-replay', message_callback=discard) == 0
    assert get_queue_size('test-replay-dl') == 0
    assert get_queue_size('test-replay') == 0


def leave(**kwargs):
    return LEAVE


def test_leave(dead_letter_message):
    assert replay_event('test-replay', message_callback=leave) == 0
    assert get_queue_size('test-replay-dl') == 1
    assert get_queue_size('test-replay') == 0
    header, event = get_message_from_queue('test-replay-dl')
    assert event.data == dead_letter_message


def nop(event):
    pass


def test_empty_queue():
    name = 'test-empty-queue'
    subscriber = Subscriber()
    subscriber.register(nop, name, ['test.empty'], dead_letter=True)
    assert replay_event('test-empty-queue') == 0
